# -*- coding: utf-8 -*-
"""
Created on Thu Mar 16 16:11:08 2023

@author: dphilippus
"""

"""
Define HEC-RAS geometry class(es) as a standardized format.
"""


from copy import deepcopy


def rs2float(rs):
    # Convert river station to float (because interpolated
    # stations are nnn.*)
    if rs[-1:] == "*":
        return float(rs[:-1])
    else:
        return float(rs)


class Geometry(object):
    # For consistency, all coordinates are adjusted so that 0 is the left
    # extreme and 0 is the minimum elevation.  However, datum and offset
    # are stored so real coordinates can be recalculated later.
    def __init__(self, coord, mann, banksta):
        # Coord: X-Y pairs [(sta, elev)]
        # Mann: X-n pairs [(sta, manning)]
        # Banksta: (left, right)
        #
        # Offset: for consistency, left bank = 0; store the offset, though,
        # to recreate later.
        # Likewise datum => minimum elevation.
        self.offset = coord[0][0]
        self.datum = min([co[1] for co in coord])
        self.coordinates = [
            (co[0] - self.offset, co[1] - self.datum) for co in coord
            ]
        self.roughness = [(mn[0] - self.offset, mn[1]) for mn in mann]
        self.banks = (banksta[0] - self.offset, banksta[1] - self.offset)

    def restore(self):
        # Recreate HEC-RAS-style coordinates (with offset and datum)
        return {
            "coordinates":
                [(co[0] + self.offset, co[1] + self.datum)
                 for co in self.coordinates],
            "roughness":
                [(mn[0] + self.offset, mn[1]) for mn in self.roughness],
            "banks":
                (self.banks[0] + self.offset, self.banks[1] + self.offset)
            }

    def update(self, geofun):
        # Update (in place) with a geometry function
        (self.coordinates, self.roughness, self.banks) = geofun(
            self.coordinates, self.roughness, self.banks)
        return self

    def adjusted(self, geofun):
        # Return updated copy with geometry function
        return deepcopy(self).update(geofun)


class Reach(object):
    # Define a Reach class to easily track datums, etc.
    def __init__(self, name, geometries):
        # Geometries should be a dictionary of {station: geometry}
        self.name = name
        self.geometries = geometries  # dictionary
        # Store both numeric value (for sorting, etc) and string value
        # for exact identification (no float errors)
        self.stations = {rs2float(x): x for x in geometries}
        self.upstream = max(self.stations)
        self.downstream = min(self.stations)
        self.re_datums()  # compute datums

    def __repr__(self):
        return "Reach %s: length %.2f units with %d cross-sections" % (
            self.name.strip(), max(self.stations) - min(self.stations),
            len(self.stations))

    def re_datums(self):
        # Recalculate datums after changing geometries
        # Make sure they are in order
        self.datums = [self.geometries[self.stations[x]].datum
                       for x in sorted(self.stations)]

    def get_sta(self, first=None, last=None):
        # Retrieve ordered, _numerical_ list of stations (i.e. float not str)
        first = min(self.stations) if first is None else first
        last = max(self.stations) if last is None else last
        return [sta
                for sta in sorted(self.stations)
                if sta >= first and sta <= last]

    def set_datums(self, delta, first=None, last=None):
        # Update datums by a specified amount
        # If first/last are not specified, will use upstream and downstream
        # stations.
        # Order is from downstream to upstream.
        # Modifies object in-place.
        to_update = [self.stations[sta] for sta in self.get_sta(first, last)]
        if len(to_update) != len(delta) and len(delta) != 1:
            raise ValueError("length of delta != number of stations")
        for ix in range(len(to_update)):
            self.geometries[to_update[ix]].datum += (
                delta[ix] if len(delta) != 1 else delta[0])
        self.re_datums()
        return self

    def adjust_datums(self, down_adj, up_adj=None, first=None, last=None):
        # Update datums from first to last.
        # If up_adj is None, update everything by down_adj.
        # Otherwise, linearly interpolate the adjustment based on the lengths.
        # Modifies and returns a copy.
        if up_adj is None:
            return self.set_datums([down_adj], first, last)
        else:
            to_update = self.get_sta(first, last)
            tot_len = max(to_update) - min(to_update)
            # For interpolation
            slope = (up_adj - down_adj) / tot_len
            lengths = [sta - to_update[0] for sta in to_update]
            delta = [down_adj + slope * dist for dist in lengths]
            return deepcopy(self).set_datums(delta, first, last)

    def set_geometry(self, geofun, first=None, last=None):
        # Apply a geometry adjustment function to selected cross-sections
        # Modifies in place.
        to_update = [self.stations[sta] for sta in self.get_sta(first, last)]
        for ud in to_update:
            self.geometries[ud] = self.geometries[ud].update(geofun)
        self.re_datums()
        return self

    def adjust_geometry(self, geofun, first=None, last=None):
        # Like set_geometry; returns a copy.
        return deepcopy(self).set_geometry(geofun, first, last)



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 16 16:10:25 2023

@author: dphilippus
"""

"""
Parse HEC-RAS geometry files into a Python-friendly structure.

HEC-RAS geometry files (i.e. xyz.g01, etc) begin with an overall description,
then junction descriptions.  These are not of interest to RaspyGeo.

Next, there is a reach description, including name, short ID, etc.  This
is followed by a long list of what appear to be cross-section coordinates.
The cross-section is identified by:
River Reach=Compton Creek   ,CC      

The cross-section geometries are of the form:
Type RM Length L Ch R = 1 ,43505   ,139,139,139
BEGIN DESCRIPTION:
Whittier Narrows Dam (D/S Face 441+12.81) u/s limit of project reach
END DESCRIPTION:
XS GIS Cut Line=5
6536310.509547571829544.635834546536310.509547571829544.63583454
6536264.334177741829598.175436616536218.158807921829651.71503868
6536218.158807921829651.71503868
Node Last Edited Time=May/03/2022 10:53:19
#Sta/Elev= 15 
   930.5  207.71   930.5  186.99  968.89  186.99  968.89   184.4  972.85  183.74
  987.95  183.74  993.19  182.43  999.75  182.43 1006.31  182.43 1011.55  183.74
 1026.65  183.74 1030.61   184.4 1030.61  186.99    1069  186.99    1069  207.71
#Mann= 6 ,-1 , 0 
   930.5    .017       0   930.5    .017       0  968.89     .15       0
  987.95    .035       0 1011.55     .15       0 1030.61    .017       0
Bank Sta=930.5,1069
XS Rating Curve= 0 ,0
XS HTab Starting El and Incr=187.49,1, 20 
XS HTab Horizontal Distribution= 5 , 5 , 5 
Exp/Cntr=0,0

The first line includes the river station (second number) and the lengths.
River station is of interest for scenarios.  Further down, after #Sta/Elev= n,
there are n coordinate pairs.  Next is #Mann= n,??,??, identifying the
number of Manning's n change points, and then trios of numbers identifying
those (station, n, ???).  Next there are the bank stations.
It appears likely that the second flag in #Mann is the type of variation,
where 0 is LOB/ROB/MC and -1 is horizontally varying.

In order to run the required analysis for scenarios, it is necessary to
retrieve roughness, geometry, and location.  From these can be computed
the datum, etc, but that is best left to hecgeo.

To write, it will be necessary to:
    - Read the whole geometry file
    - Identify and edit relevant cross-sections
    - Correct Sta/Elev and Mann headers and data
    - Write the new file
"""


from RaspyGeo.hecgeo import Geometry, Reach


def mk_name(raw):
    return ",".join([i.strip() for i in raw.split(",")])


def first_line(text):
    return text[:text.find("\n")]


def rest_lines(text):
    return text[(text.find("\n")+1):]


def make_geo(text):
    # Cross-section text => Geometry object.
    # Geometry arguments: coordinates [(x, y)], roughness [(x, mann)],
    # bank stations (left, right)
    statext = rest_lines(text[text.find("#Sta/Elev="):text.find("#Mann=")])
    # Need to work with remaining text for Manning's.
    rtx = rest_lines(text[text.find("#Mann="):])
    manntext = rtx[:rtx.find("=")]
    # Weirdness: cut off at the _last_ newline
    manntext = manntext[:(len(manntext) - manntext[::-1].find("\n")-1)]
    banktext = first_line(text[text.find("Bank Sta="):]).split("=")[1]
    # Process station data
    # Excluding very long chunks: sometimes when one is too long it spills
    # over; this tends to occur in high-density survey data, so excluding
    # one point should not be a huge problem.
    stalist = [x for x in statext.split(" ") if x and len(x) <= 7]
    sta = [(float(stalist[i]), float(stalist[i+1]))
           for i in range(0, len(stalist)-1, 2)]
    # Process Manning's data
    mannlist = [x for x in manntext.split(" ") if x]
    mann = [(float(mannlist[i]), float(mannlist[i+1]))
            for i in range(0, len(mannlist)-2, 3)]
    # Proces bank stations
    banklist = banktext.split(",")
    banks = (float(banklist[0]), float(banklist[1]))
    return Geometry(sta, mann, banks)


def get_rs(block):
    # First line will be `1 ,43505   ,139,139,139` or similar
    return first_line(block).split(",")[1].strip()


def sep_inner(name, text):
    # Reach-specific text => Reach
    # Reach-specific text is between two pairs of "River Reach=", excluding
    # that particular row.
    return Reach(name, {
        get_rs(x): make_geo(rest_lines(x))
        for x in text.split("Type RM Length L Ch R = ")[1:]
        # Filter: exclude XSes missing data (e.g. bridges)
        if rest_lines(x).find("Bank Sta=") >= 0
        and rest_lines(x).find("#Sta/Elev=") >= 0
        and rest_lines(x).find("#Mann=") >= 0
        })


def parse(file):
    # Read the file path, then separate it into
    # {reach: fn(name, text)}
    with open(file, "r") as f:
        raw = f.read()
    # Edge cases go here
    # Prevents channel modification alternatives from causing problems
    raw = raw.split("CM Alternative")[0]
    return {
        mk_name(first_line(x)):
            sep_inner(mk_name(first_line(x)), rest_lines(x))
        for x in raw.split("River Reach=")[1:]
        }



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 24 13:31:10 2023

@author: dphilippus
"""

"""
Write Python Reach objects back to HEC-RAS.

HEC-RAS geometry files (i.e. xyz.g01, etc) begin with an overall description,
then junction descriptions.  These are not of interest to RaspyGeo.

Next, there is a reach description, including name, short ID, etc.  This
is followed by a long list of what appear to be cross-section coordinates.
The cross-section is identified by:
River Reach=Compton Creek   ,CC      

The cross-section geometries are of the form:
Type RM Length L Ch R = 1 ,43505   ,139,139,139
BEGIN DESCRIPTION:
Whittier Narrows Dam (D/S Face 441+12.81) u/s limit of project reach
END DESCRIPTION:
XS GIS Cut Line=5
6536310.509547571829544.635834546536310.509547571829544.63583454
6536264.334177741829598.175436616536218.158807921829651.71503868
6536218.158807921829651.71503868
Node Last Edited Time=May/03/2022 10:53:19
#Sta/Elev= 15 
   930.5  207.71   930.5  186.99  968.89  186.99  968.89   184.4  972.85  183.74
  987.95  183.74  993.19  182.43  999.75  182.43 1006.31  182.43 1011.55  183.74
 1026.65  183.74 1030.61   184.4 1030.61  186.99    1069  186.99    1069  207.71
#Mann= 6 ,-1 , 0 
   930.5    .017       0   930.5    .017       0  968.89     .15       0
  987.95    .035       0 1011.55     .15       0 1030.61    .017       0
Bank Sta=930.5,1069
XS Rating Curve= 0 ,0
XS HTab Starting El and Incr=187.49,1, 20 
XS HTab Horizontal Distribution= 5 , 5 , 5 
Exp/Cntr=0,0

The first line includes the river station (second number) and the lengths.
River station is of interest for scenarios.  Further down, after #Sta/Elev= n,
there are n coordinate pairs.  Next is #Mann= n,??,??, identifying the
number of Manning's n change points, and then trios of numbers identifying
those (station, n, ???).  Next there are the bank stations.
It appears likely that the second flag in #Mann is the type of variation,
where 0 is LOB/ROB/MC and -1 is horizontally varying.

For writing, an existing file will be taken as a template, and cross-section
portions corrected only as needed.

This requires identifying cross-section portions and editing #Sta/Elev through
to Bank Sta.  In that portion the numbers of Sta/Elev, the full #Mann string,
and the Bank Sta will need to be corrected.  The coordinate pairs for geometry
and roughness will also need to be set.

Geometry is ten fixed-width, space-delimited columns (two decimal place
precision).  The columns alternate X and Y.  Each column is eight characters
wide, filled with spaces at the beginning.

Manning's columns are also eight characters wide.  There are nine columns,
in trios of location, roughness (with no leading zero, up to three decimals),
and then a zero.

#Sta/Elev= is followed by three characters (possibly leading spaces) as an
integer, followed by a trailing space.

#Mann= is followed by ten characters.  For our purposes, it is ` N ,-1 , 0 `
where N is the number of roughness points.
Bank Sta=left,right (floats, two decimals) with no special formatting.

In general, floats should use the minimum required number of decimal places,
and no more than two (three for Manning's).
"""


from RaspyGeo.parse_geo import first_line, rest_lines, get_rs, parse, mk_name


def fmt_num(x):
    return "% 8.2f" % x


def fmt_mann(x):
    return "% 8.3f" % x


def fmt_nsta(x):
    if x > 99:
        return str(x)
    else:
        return "% 2d" % x


def flatten(ls):
    return [i for x in ls for i in x]


def blockify(ls, N):
    return '\n'.join([
        ''.join(ls[k:min(k+N, len(ls))])
        for k in range(0, len(ls), N)
        ])


def fmt_name(rchn):
    # Reach name -> HEC-RAS format: river and reach 16 characters,
    # left- and right-aligned, space-padded, comma separated
    ls = rchn.split(",")
    return ls[0] + " " * (16 - len(ls[0])) + "," + " " * (16 - len(ls[1])) +\
        ls[1]


def coordinates(coords):
    # Convert coordinates [(X,Y)] to the geometry block (10 fixed-width
    # columns)
    # with header #Sta/Elev= N
    allcol = [fmt_num(x) for x in flatten(coords)]
    block = blockify(allcol, 10) if len(allcol) > 10 else ''.join(allcol)
    return "#Sta/Elev= %d \n%s" % (len(coords), block)


def mann(ro):
    allcol = flatten([(fmt_num(x[0]),
                       fmt_mann(x[1]),
                       '       0') for x in ro])
    block = blockify(allcol, 9)
    header = '#Mann=%s%d ,-1 , 0 ' % (
        '' if len(ro) > 9 else ' ',
        len(ro))
    return '%s\n%s' % (header, block)


def banksta(banks):
    return 'Bank Sta=%.2f,%.2f' % banks


def edit_block(original,
               newgeo):
    # Takes a Geometry object as an input in addition to the original
    # text block.
    geos = newgeo.restore()
    # Find Manning's block, since the next block is not fixed.
    # Each Manning's roughness occupies exactly 24 characters,
    # plus one for each time there are more than 3.
    # The title block has exactly 18 characters (including newline).
    mix = original.find('#Mann')
    mann_n = int(original[mix:][:20]
                 .split("=")[1]
                 .split(',')[0]
                 .strip())
    nnewln = 1 + int((mann_n - 1) / 3)
    nch = 24 * mann_n + nnewln + 18
    # origina[endman:] starts right after Manning's (without leading newline)
    endman = mix + nch
    # Next, identify the bank station block
    bankix = original.find('Bank Sta=')
    endbank = bankix + original[bankix:].find('\n') + 1
    # Assemble slices
    slices = [
        # #Sta/Elev is easy.
        original[:(original.find('#Sta/Elev')-1)],
        coordinates(geos["coordinates"]),
        mann(geos["roughness"]),
        # Middle bit (may be '', hence filtering)
        original[endman:bankix],
        banksta(geos["banks"]),
        original[endbank:]
        ]
    return "\n".join([slc for slc in slices if slc != ''])


def proc_reach(name, text, reaches):
    # The processor function gets a name, which is the reach name
    # (first line), and the remaining reach text.
    # It should then be possible to loop through XS blocks,
    # edit them accordingly, and rebuild the file.
    # Reaches (e.g. returned by parse) have the `name` values as their
    # keys, so that makes it easy.  Cross-sections have get_rs(block), for
    # each block, as their key within Reach.geometries, also conveniently.
    rch = reaches[mk_name(name)]
    sep = "Type RM Length L Ch R = "
    chunks = text.split(sep)
    header = chunks[0]
    return name + "\n" + sep.join([header] + [
        edit_block(block, rch.geometries[get_rs(block)])
        # If it is not in the geometry (e.g. a bridge), just return the
        # old one.
        if get_rs(block) in rch.geometries else block
        for block in chunks[1:]
        ])


def read_write(file, reaches, out=None):
    # Read the file path, then separate it into
    # {reach: fn(name, text)}
    out = out if out is not None else file
    with open(file, "r") as f:
        raw = f.read()
    with open(file + ".bak", "w") as f:
        f.write(raw)
    chunks = raw.split("River Reach=")
    data = "River Reach=".join([chunks[0]] + [
        proc_reach(first_line(x), rest_lines(x), reaches)
        for x in chunks[1:]])
    with open(out, "w") as f:
        f.write(data)


def read_modify(file, modfns, out=None):
    # modfns => {reach name: f(Reach)} where f modifies the Reach as desired.
    reaches = parse(file)
    newrch = {rch: modfns[rch](reaches[rch])
              if rch in modfns else reaches[rch]
              for rch in reaches}
    read_write(file, newrch, out)



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 27 10:07:44 2023

@author: dphilippus
"""

"""
Iterate through scenarios and retrieve results.
"""

from RaspyGeo.write_geo import read_modify
from RaspyGeo.geofun import set_afp, set_lfc
from raspy_auto import API, Ras


"""
Example scenarios:
def r1l(geo):
    return geo  .adjust_datums(0, 100, 500, 2000) \
                .adjust_geometry(
                    set_afp(10, 3, 4, lambda x: x/2,
                            2, 0.123, 0.246, 0.035, 0.017))


def r1u(geo):
    return geo.adjust_datums(200, 250)


# The argument to `run`
mods = {'RiverOne        ,Upper           ': r1u,
 'RiverOne        ,Lower           ': r1l}

Test proj has 10 profiles.
"""


cols = "Scenario,ID,River,Reach,RS,Q,shear.lob,shear.mc,shear.rob,\
vel.lob,vel.mc,vel.rob,depth.lob,depth.mc,depth.rob\n"


def row_join(row):
    return ",".join([str(r) for r in row])


def twovalfix(vals, maxC=False, av=False):
    # Sometimes HEC-RAS will return just two values.  This fixes it to
    # distribute them across the overbanks and MC.
    # If maxC (max center), the max value goes in the MC, and the missing
    # value is set to 0.
    # If av (average), then all three are set to the mean.
    # If both are false, then it just returns three NAs.
    if len(vals) == 3:
        return vals
    if len(vals) == 1:
        return [0, vals[0], 0]
    if maxC:
        if vals[0] > vals[1]:  # larger is left
            return [0] + vals
        else:  # larger is right
            return vals + [0]
    if av:
        mean = sum(vals)/len(vals)
        return [mean, mean, mean]
    else:
        return ["NA", "NA", "NA"]


def loc_data(ras, nprof, loc, scenario):
    # loc -> [identifier, river, reach, rs]
    # Returns [formatted row according to `cols`] for a single location
    # flow_data: {profile: SimData}, where SimData has
    # velocity, maxDepth, flow, shear as lists of [l, c, r]
    ident = loc[0]
    riv = loc[1].strip()
    rch = loc[2].strip()
    rs = loc[3].strip()
    flow_data = ras.data.allFlowDist(riv,
                                     rch,
                                     rs,
                                     nprof)
    return [row_join(
        [scenario, ident, riv, rch, rs,
         sum(fd.flow)] +
        twovalfix(fd.shear, av=True) +
        twovalfix(fd.velocity, maxC=True) +
        twovalfix(fd.maxDepth, maxC=True)
        ) for fd in flow_data.values()]


def scenario_data(ras, nprof, locations, scenario):
    # locations -> [[identifier, river, reach, rs]]
    # nprof -> number of flow profiles
    # Returns [formatted row according to `cols`]
    return [
        row for loc in locations
        for row in loc_data(ras, nprof, loc, scenario)
        ]


def run(projPath, ingeo, outgeo, outfile, locations, nprof, scenarios,
        which="507"):
    # projPath -> project location
    # locations -> [[identifier, river, reach, rs]] for data retrieval
    # Loop through scenarios, set geometry, run simulation, and retrieve data.
    # Scenarios should be a dictionary with labels.  These are used for
    # writing.
    # `outfile` will be overwritten.
    ras = API(Ras(projPath, which=which))
    with open(outfile, "w") as f:
        f.write(cols)
        for scen in scenarios:
            read_modify(ingeo, scenarios[scen], outgeo)
            ras.ops.openProject(projPath)
            ras.ops.compute()
            f.write("\n".join(
                scenario_data(ras, nprof, locations, scen)) + "\n")
